package ru.windcorp.turretlang;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

public class TLConfiguration {
	private final File file;
	protected int failed = 0;
	
	protected List<TLVariable> vars = new ArrayList<TLVariable>();
	
	TLConfiguration(File file) throws IOException, TLException {
		this.file = file.getAbsoluteFile();
		reload();
	}
	
	/**
	 * Returns the absolute {@link File} that is the source of this configuration.
	 * @return The source.
	 * */
	public final File getSource() {
		return file;
	}
	
	/**
	 * Returns the number of lines failed to load during the last reload of the configuration.
	 * @return The number of failed lines.
	 * */
	public final int getFailed() {
		return failed;
	}
	
	/**
	 * Returns a {@link TLVariable} by the specified name, ignoring case.
	 * @return A TLVariable with the specified name, or null if no such variable exists.
	 * @param name The name to search against. If null is provided, "null" is used instead.
	 * */
	public TLVariable getVariable(String name) {
		if (name == null) name = "null";
		for (TLVariable v : vars) {
			if (v.getName().equalsIgnoreCase(name)) return v;
		}
		return null;
	}
	
	/**
	 * Returns a {@link TLVariable} by the specified name, ignoring case.
	 * @return A TLVariable with the specified name, or the specified one if no such variable exists.
	 * @param name The name to search against. If null is provided, "null" is used instead.
	 * @param def The default value used in case no variable is found.
	 * */
	public TLVariable getVariable(String name, TLVariable def) {
		TLVariable result = getVariable(name);
		if (result == null) return def; return result;
	}
	
	/**
	 * Returns the String representation of the value of the variable by the specified name, ignoring case.
	 * @return A String representation of the value of the variable by the specified name, or null if no such variable exists.
	 * @param name The name to search against. If null is provided, "null" is used instead.
	 * */
	public String getValue(String name) {
		if (name == null) name = "null";
		for (TLVariable v : vars) {
			if (v.getName().equalsIgnoreCase(name)) return v.getValue();
		}
		return null;
	}
	
	/**
	 * Returns the String representation of the value of the variable by the specified name, ignoring case.
	 * @return A String representation of the value of the variable by the specified name, or the specified one if no such variable exists.
	 * @param name The name to search against. If null is provided, "null" is used instead.
	 * @param def The default value used in case no variable is found.
	 * */
	public String getValue(String name, String def) {
		String result = getValue(name);
		if (result == null) return def; return result;
	}
	
	/**
	 * Returns a {@link List} of all the variables with the specified value. Case-sensitive.
	 * @return A List of all the variables with the specified value.
	 * @param value The name to search against. If null is provided, "null" is used instead.
	 * */
	public List<TLVariable> getByValue(String value) {
		List<TLVariable> result = new ArrayList<TLVariable>();
		if (value == null) value = "null";
		
		for (TLVariable v : vars) {
			if (v.getValue().equals(value)) result.add(v);
		}
		
		return result;
	}
	
	/**
	 * Checks the existence of a variable with such name.
	 * @return True if such a variable exists, false otherwise.
	 * @param name The name to check against. If null is provided, "null" is used instead.
	 * */
	public boolean exists(String name) {
		return getVariable(name) != null;
	}
	
	/**
	 * Checks the existence of the provided variable.
	 * @return True if such a variable exists, false otherwise.
	 * @param var The variable to check against. If null is provided, "null" is used instead.
	 * */
	public boolean exists(TLVariable var) {
		return vars.contains(var);
	}
	
	/**
	 * Returns all the variables in the configuration as a {@link List}.
	 * @return All the variables in this configuration.
	 * */
	public List<TLVariable> contents() {
		return vars;
	}
	
	/**
	 * Returns the total amount of variables contained in the configuration.
	 * @return The total amount of variables in this configuration.
	 * */
	public int size() {
		return vars.size();
	}
	
	/**
	 * Reloads the configuration from the {@link TLConfiguration#getSource() source}.
	 * In case of an exception the contents do not change.
	 * @throws TLException If the file could not be red correctly or doesn't exist anymore.
	 * @return The amount of failed declarations.
	 * @throws IOException If an IO error occurs.
	 * @see TurretLang#load(File)
	 * */
	public int reload() throws IOException, TLException {
		List<String> declrs = TurretLang.loadVars(file);
		
		vars.clear();
		failed = 0;
		for (String d : declrs) {
			try {
				if (!d.startsWith("*") || d.equals(""))
					vars.add(new TLVariable(d));
			} catch (TLException e) {
				failed++;
			}
		}
		
		return failed;
	}
}
