package ru.windcorp.turretlang;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Map;

import ru.windcorp.turretlang.saveable.TLSaveableException;
import ru.windcorp.turretlang.util.FileIOUtil;

/**
 * Represents a variable (a junction of a name and a value)
 * */
public class TLVariable {
	protected String declaration, name, value;
	
	/**
	 * A class that "wraps" the boolean primitive type so that it can be null.
	 * */
	public final class NullableBoolean {
		private final int value;
		
		public static final int TRUE = 0, FALSE = 1, NULL = 2;
		
		public NullableBoolean(int value) {
			this.value = value;
		}
		
		/**
		 * Intends to convert self into a boolean. If it is null, it will throw an UnsupportedOperationException.
		 * @return The boolean value of this NullableBoolean, if such exists.
		 * @throws UnsupportedOperationException If the current value is null.
		 * @see NullableBoolean#convertToWrappedBool()
		 * */
		public boolean convert() throws UnsupportedOperationException {
			if (value == 0) return true;
			if (value == 1) return false;
			throw new UnsupportedOperationException("The NullableBoolean equals null and could not be converted to native boolean");
		}
		
		/**
		 * Checks if the current value is null.
		 * @return True, if this NullableBoolean represents a null, false otherwise.
		 * */
		public boolean isNull() {
			return value != 0 && value != 1;
		}
		
		/**
		 * A convenience method to get a {@link Boolean} object, see {@link NullableBoolean#convert()} for details.
		 * @return The {@link Boolean} value of this NullableBoolean, if such exists.
		 * @throws UnsupportedOperationException If the current value is null.
		 * */
		public Boolean convertToWrappedBool() throws UnsupportedOperationException {
			return Boolean.valueOf(convert());
		}
		
		@Override public String toString() {
			try {
				return convertToWrappedBool().toString();
			} catch (UnsupportedOperationException e) {
				return "null";
			}
		}
	}
	
	/**
	 * Creates a new variable from its declaration.
	 * @param declaration The variable's declaration.
	 * @throws TLException If the declaration provided is invalid. 
	 * */
	public TLVariable(String declaration) throws TLException {
		this.declaration = declaration;
		this.name = TurretLang.getVariableName(declaration);
		this.value = TurretLang.getVariableValue(declaration);
	}
	
	/**
	 * Gets the original declaration of the variable.
	 * @return The declaration.
	 * */
	public String getDeclaration() {
		return declaration;
	}
	
	/**
	 * Gets the name of the variable.
	 * @return The name.
	 * */
	public String getName() {
		return name;
	}
	
	/**
	 * Gets the original value of the variable as a {@link String}.
	 * @return The string value.
	 * */
	public String getValue() {
		return value;
	}
	
	/**
	 * Returns the variable value as a boolean. In case it is not a boolean, it will return false.
	 * For most cases {@link TLVariable#getValueAsTrueBoolean()} should be used to determine if the 'false' value is set so or is simply absent.
	 * @return True if and only if the value equals "true" (ignoring case), false otherwise.
	 * @see TLVariable#getValueAsTrueBoolean()
	 * @see TLVariable#getValueAsBooleanOrDefault(boolean)
	 * */
	public boolean getValueAsBoolean() {
		return Boolean.getBoolean(value);
	}
	
	/**
	 * Returns the variable value as a {@link NullableBoolean}. Unlike {@link TLVariable#getValueAsBoolean()},
	 * it will return a null-filled object (not null) in case a boolean is not indicated. <p>
	 * For example, "tRue" will return a NullableBoolean that returns true when {@link NullableBoolean#convert() converted},
	 * "FALSE" in the same situation will result in false while
	 * "yes" or "blah-blah-BLAH" will cause an exception if converted.
	 * @return A NullableBoolean representing this boolean.
	 * @see NullableBoolean#convert()
	 * */
	public NullableBoolean getValueAsTrueBoolean() {
		if (value.equalsIgnoreCase("true")) return new NullableBoolean(0);
		if (value.equalsIgnoreCase("false")) return new NullableBoolean(1);
		return new NullableBoolean(2);
	}
	
	/**
	 * Returns the variable value as a boolean.
	 * @return The boolean value or, if not possible to convert, the specified value.
	 * @param value The value to return if the variable's value is not a boolean.
	 * */
	public boolean getValueAsBooleanOrDefault(boolean value) {
		if (this.getValueAsTrueBoolean().isNull()) return value; return this.getValueAsTrueBoolean().convert();
	}
	
	/**
	 * Returns the variable value as a byte.
	 * @return The byte representation of the value.
	 * @throws NumberFormatException If the value is not numeric.
	 * @throws ClassCastException If the value is not a byte.
	 * */
	public byte getValueAsByte() throws NumberFormatException, ClassCastException {
		long l = this.getValueAsLong();
		if (l < Byte.MIN_VALUE || l > Byte.MAX_VALUE)
			throw new ClassCastException(l + " is too big or to small to be a byte");
		return (byte) l;
	}
	
	public byte getValueAsByteOrDefault(byte value) {
		try {
			return getValueAsByte();
		} catch (Exception e) {return value;}
	}
	
	/**
	 * Returns the variable value as a short.
	 * @return The short representation of the value.
	 * @throws NumberFormatException If the value is not numeric.
	 * @throws ClassCastException If the value is not a short.
	 * */
	public short getValueAsShort() throws NumberFormatException, ClassCastException {
		long l = this.getValueAsLong();
		if (l < Short.MIN_VALUE || l > Short.MAX_VALUE)
			throw new ClassCastException(l + " is too big or to small to be a short");
		return (short) l;
	}
	
	public short getValueAsShortOrDefault(short value) {
		try {
			return getValueAsShort();
		} catch (Exception e) {return value;}
	}
	
	/**
	 * Returns the variable value as an int.
	 * @return The int representation of the value.
	 * @throws NumberFormatException If the value is not numeric.
	 * @throws ClassCastException If the value is not an int.
	 * */
	public int getValueAsInt() throws NumberFormatException, ClassCastException {
		long l = this.getValueAsLong();
		if (l < Integer.MIN_VALUE || l > Integer.MAX_VALUE)
			throw new ClassCastException(l + " is too big or to small to be an integer (number)");
		return (int) l;
	}
	
	public int getValueAsIntOrDefault(int value) {
		try {
			return getValueAsInt();
		} catch (Exception e) {return value;}
	}
	
	/**
	 * Returns the variable value as a long.
	 * @return The long representation of the value.
	 * @throws NumberFormatException If the value is not numeric.
	 * */
	public long getValueAsLong() throws NumberFormatException {
		return Long.parseLong(value);
	}
	
	public long getValueAsLongOrDefault(long value) {
		try {
			return getValueAsLong();
		} catch (Exception e) {return value;}
	}
	
	/**
	 * Returns the variable value as a float.
	 * @return The float representation of the value.
	 * @throws NumberFormatException If the value is not numeric.
	 * @throws ClassCastException If the value is not an float.
	 * */
	public float getValueAsFloat() throws NumberFormatException, ClassCastException {
		double d = this.getValueAsDouble();
		if (d < Float.MIN_VALUE || d > Float.MAX_VALUE)
			throw new ClassCastException(d + " is too big or to small to be a float (a number with a decimal part)");
		return (int) d;
	}
	
	public float getValueAsFloatOrDefault(float value) {
		try {
			return getValueAsFloat();
		} catch (Exception e) {return value;}
	}
	
	/**
	 * Returns the variable value as a double.
	 * @return The double representation of the value.
	 * @throws NumberFormatException If the value is not numeric.
	 * */
	public double getValueAsDouble() throws NumberFormatException {
		return Double.parseDouble(value);
	}
	
	public double getValueAsDoubleOrDefault(double value) {
		try {
			return getValueAsDouble();
		} catch (Exception e) {return value;}
	}
	
	/**
	 * Returns the variable as a String array, splitting it with ';'.<p>
	 * E.g. "red;blue;;green;yellow;" returns {"red", "blue", "", "green", "yellow"}.
	 * @return A String array represented by this value.
	 * @see String#split(String)
	 * */
	public String[] getValueAsArray() {
		return value.split(";");
	}
	
	/**
	 * Returns the first character of the value.<p>
	 * E.g. "portal" returns 'p'.
	 * @return The first character of the value.
	 * @throws IndexOutOfBoundsException If the string is "".
	 * */
	public char getValueAsChar() throws IndexOutOfBoundsException {
		return value.charAt(0);
	}
	
	public char getValueAsCharOrDefault(char value) {
		try {
			return this.value.charAt(0);
		} catch (Exception e) {
			return value;
		}
	}
	
	/**
	 * Returns the map represented by the value following this syntax:<p>
	 * "key0@val0;key1@val1" will result in a map with two keys: "key0" and "key1" and two corresponding values "val0" and "val1".
	 * @return A map represented by this value.
	 * @throws IllegalStateException If some key does not have a pair, e.g. in "key0@val0;<b>key1</b>;key2@val2"
	 * key1 will cause an error.
	 * */
	public Map<String, String> getValueAsMap() throws IllegalStateException {
		Map<String, String> result = new HashMap<String, String>();
		String[] items = value.split(";");
		
		for (String s : items) {
			String[] vk = s.split("@", 2);
			if (vk.length < 2)
				throw new IllegalStateException("This variable does not represent a map");
			result.put(vk[0], vk[1]);
		}
		
		return result;
	}
	
	public Map<String, String> getValueAsMapOrDefault(Map<String, String> value) {
		try {
			return getValueAsMap();
		} catch (Exception e) {
			return value;
		}
	}
	
	/**
	 * Checks if this value is an exact number. The value is considered to be so if it can be parsed to a long.
	 * @return True if the value is an exact number, false otherwise.
	 * @see TLVariable#isNumeric()
	 * */
	public boolean isExactNumeric() {
		try {
			getValueAsLong();
			return true;
		} catch (Exception e) {
			return false;
		}
	}
	
	/**
	 * Checks if this value is numeric. The value is considered numeric if it can be parsed to a double.
	 * @return True if the value is numeric, false otherwise.
	 * @see TLVariable#isExactNumeric()
	 * */
	public boolean isNumeric() {
		try {
			getValueAsDouble();
			return true;
		} catch (Exception e) {
			return false;
		}
	}
	
	/**
	 * Returns the value as a formatted Minecraft message, replacing '&#38;' by a paragraph sign were it marks a chat color.
	 * <i>Notice that TurretLang API does not use any Bukkit libraries so this will cause errors if ChatColor updates.</i>
	 * @return A String with chat colors.
	 * */
	public String getValueWithChatColors() {
		String result = getValue();
		char[] codes = new char[] {
				'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f',
				'l', 'o', 'n', 'm', 'k', 'r'
		};
		
		for (char c : codes) {
			result.replaceAll("&" + c, "\u00a7" + c);
		}
		
		return result;
	}
	
	/**
	 * Returns the value as an Object. Note that the represented Object must have several properties,
	 * such as a public constructor with argument of type "String[]" with no exceptions thrown, even in case no arguments are expected.<p>
	 * The syntax is different from arrays and maps: it looks like
	 * "<code>package.className*arg0*arg1*arg2...</code>" or "<code>~null</code>" (to indicate a null object).<p>
	 * When instantiated the arguments are passed in the order of their declaration.
	 * @return A new instance of the represented Object or null if indicated so.
	 * @throws TLSaveableException A wrap for all exceptions that may be thrown during reflection operations;
	 * includes the argument declaration.
	 * @see TLVariable#getValueAsObjectOrDefault(Object, boolean)
	 * @see TLVariable#getValueAsObjectOrDefault(Object)
	 * */
	public Object getValueAsObject() throws TLSaveableException {
		String d = getValue();
		
		if (d.equals("~null")) return null;
		
		String[] dparts = d.split("*", 2);
		if (dparts.length <= 1) throw new TLSaveableException(d + " is not an Object declaration", d);
		
		try {
			Class<?> cl = Class.forName(dparts[0]);
			Constructor<?> c = cl.getConstructor(String[].class);
			
			String[] args = dparts[1].split("*");
			return c.newInstance((Object[]) args);
			
		} catch (ClassNotFoundException e) {
			throw new TLSaveableException(dparts[0] + " is an unknown classname", e, d);
		} catch (NoSuchMethodException e) {
			throw new TLSaveableException("Constructor \"public " + dparts[0] + "(String[] args)\" is absent", e, d);
		} catch (SecurityException e) {
			throw new TLSaveableException("Constructor \"" + dparts[0] + "(String[] args)\" is not public", e, d);
		} catch (InstantiationException e) {
			throw new TLSaveableException(
					"Constructor \"public " + dparts[0] + "(String[] args)\" is not avaliable (the class cannot be instantiated)", e, d);
		} catch (IllegalAccessException e) {
			throw new TLSaveableException("Constructor \" " + dparts[0] + "(String[] args)\" is not public", e, d);
		} catch (IllegalArgumentException e) {
			throw new TLSaveableException(
					"Constructor \"public " + dparts[0] + "(String[] args)\" indicated String[] as an illegal argument", e, d);
		} catch (InvocationTargetException e) {
			throw new TLSaveableException(
					"Constructor \"public " + dparts[0] + "(String[] args)\" has thrown an exception ("
					+ e.getCause().getClass().getName() + ")", e, d);
		}
	}
	
	/**
	 * Returns the value as an Object. See {@link TLVariable#getValueAsObject()} for conditions that the Object must comply.
	 * @return The Object represented by this variable or the given value if any exception is thrown.
	 * @param value The default value to return in case the instantiation fails.
	 * @param replaceNull Indicates whether to consider null appropriate or not, where true will make the method return
	 * the default value in case the represented Object is null.
	 * @see TLVariable#getValueAsObject()
	 * */
	public Object getValueAsObjectOrDefault(Object value, boolean replaceNull) {
		try {
			Object obj = getValueAsObject();
			if (obj == null && replaceNull)
				return value;
			return obj;
		} catch (TLSaveableException e) {
			return value;
		}
	}
	
	/**
	 * Returns the value as an Object. See {@link TLVariable#getValueAsObject()} for conditions that the Object must comply.
	 * Invocation of this method is actually equal to {@link TLVariable#getValueAsObjectOrDefault(Object, boolean)
	 * TLVariable.getValueAsObjectOrDefault(value, false)}.
	 * @return The Object represented by this variable or the given value if any exception is thrown.
	 * @param value The default value to return in case the instantiation fails.
	 * @see TLVariable#getValueAsObject()
	 * */
	public Object getValueAsObjectOrDefault(Object value) {
		return this.getValueAsObjectOrDefault(value, false);
	}
	
	/**
	 * Returns the file by the path represented by the value. If any error occurs, null is returned.
	 * @return The file instance or null if the file is unreachable ({@link SecurityException}).
	 * @param absolute If true, the method will try to return the file in its absolute form, otherwise it may be abstract.
	 * */
	public File getValueAsFile(boolean absolute) {
		try {
			File f = new File(getValue());
			
			if (absolute) {
				f = f.getAbsoluteFile();
			}
			
			return f;
		} catch (Exception e) {
			return null;
		}
	}
	
	/**
	 * Returns the file by the path represented by the value. If any error occurs, null is returned.
	 * @return The file instance or null if the file is unreachable ({@link SecurityException}).
	 * @see TLVariable#getValueAsFile(boolean)
	 * */
	public File getValueAsFile() {
		return getValueAsFile(false);
	}
	
	/**
	 * Returns the file by the path represented by the value only if it exists, otherwise null.
	 * If any error occurs, null is returned, too.
	 * @return An existing file represented by this variable or null if it is unreachable or does not exist.
	 * @param checkReadable Whether to check readability of the file. If true, an existing but unreadable file will return null,
	 * otherwise the file is returned even if it is not readable.
	 * */
	public File getValueAsExistingFile(boolean checkReadable) {
		File f = getValueAsFile(true);
		
		try {
			if (f.exists()) {
				if (checkReadable)
					if (f.canRead()) return f;
					else return null;
				return f;
			} return null;
		} catch (Exception e) {
			return null;
		}
	}
	
	/**
	 * Returns the file by the path represented by the value. If no such file exists yet, it will be created and,
	 * if content is not null, data will be written to it.
	 * @return The existing file.
	 * @param content Data to write to the file if it is created. If null nothing is written.
	 * @throws IOException If any IO error occurs.
	 * @throws SecurityException If any security error occurs.
	 * */
	public File getValueAsCreatedFile(String content) throws IOException, SecurityException {
		File f = getValueAsFile(true);
		
		if (!f.exists()) {
			if (!f.getParentFile().exists()) {
				f.getParentFile().mkdirs();
			}
			
			f.createNewFile();
			
			
		}
		
		if (content != null) {
			FileIOUtil.write(f, content);
		}
		
		return f;
	}
}
