package ru.windcorp.turretlang;

import java.io.File;
import java.io.IOException;
import java.util.List;

import ru.windcorp.turretlang.util.FileIOUtil;

/**
 * The main class of Turret Language, it is used for general work with the variables.
 * */
public final class TurretLang {
	
	private TurretLang(){}
	
	/**
	 * Constant for holding the char used to separate the variable name and its value in the declaration.
	 * */
	public static final char SEPARATOR = ':';
	
	/**
	 * Version of Turret Language
	 * */
	public static final float VERSION = 1.1f;
	
	/**
	 * Gets the variable value provided in the declaration. <p>
	 * E.g. declaration "java:minecraft" will return "minecraft".
	 * @return The variable value.
	 * @param declaration The declaration, e.g. "java:minecraft"
	 * @throws TLException If the declaration is invalid.
	 * @throws IllegalArgumentException If the declaration is null.
	 * */
	public static final String getVariableValue(String declaration) throws TLException {
		assert declaration != null;
		
		String[] parts = declaration.split(SEPARATOR + "", 2);
		if (parts.length < 2) throw new TLException("Failed to process a variable: \"" + declaration + "\" is not a variable declaration");
		
		return parts[1];
	}
	
	/**
	 * Gets the variable name provided in the declaration. <p>
	 * E.g. declaration "turret:danger" will return "turret".
	 * @return The variable name.
	 * @param declaration The declaration, e.g. "turret:danger"
	 * @throws TLException If the declaration is invalid.
	 * @throws IllegalArgumentException If the declaration is null.
	 * */
	public static final String getVariableName(String declaration) throws TLException {
		assert declaration != null;
		
		String[] parts = declaration.split(SEPARATOR + "", 2);
		if (parts.length < 2) throw new TLException("Failed to process a variable: \"" + declaration + "\" is not a variable declaration");
		
		return parts[0];
	}
	
	/**
	 * Creates a String declaration for for the given variable name and the value. <p>
	 * E.g. createDeclaration("exception", "stupid") returns "exception:stupid";
	 * @return The declaration of given parameters.
	 * @param name The name of the variable.
	 * @param value The value of the variable.
	 * @throws IllegalArgumentException If some argument is null.
	 * @see TurretLang#declare(String, String)
	 * */
	public static final String createDeclaration(String name, String value) {
		assert name != null && value != null;
		return name + SEPARATOR + value;
	}
	
	/**
	 * Creates a {@link TLVariable} representation for for the given variable name and the value. <p>
	 * E.g. declare("life", "42") returns "life:42";
	 * @return The declaration of given parameters, as a {@link TLVariable} object.
	 * @param name The name of the variable.
	 * @param value The value of the variable.
	 * @throws IllegalArgumentException If some argument is null.
	 * @see TurretLang#createDeclaration(String, String)
	 * */
	public static final TLVariable declare(String name, String value) {
		try {
			return new TLVariable(createDeclaration(name, value));
		} catch (TLException e) {
			return null;
		}
	}
	
	/**
	 * Loads a file into a {@link TLConfiguration}.
	 * @return The TLConfiguration object representing the given file.
	 * @param source The file to load from.
	 * @throws IOException If an IO exception occurs.
	 * @throws TLException If the file is not available or is not a Turret Language file.
	 * */
	public static final TLConfiguration load(File source) throws IOException, TLException {
		return new TLConfiguration(source);
	}
	
	static final List<String> loadVars(File source) throws IOException, TLException {
		assert source != null;
		if (!source.canRead()) throw new TLException("Failed to load configuration from \"" + source.getAbsolutePath() + "\": could not read file", new IOException("File not readable"));
		
		return FileIOUtil.lines(source);
	}
	
	/**
	 * Checks if the given string is a valid variable declaration. 
	 * @return True if the string is a declaration, false otherwise.
	 * @param string The string to check.
	 * */
	public static final boolean isDeclaration(String string) {
		assert string != null;
		
		try {
			getVariableValue(string);
			return true;
		} catch (TLException e) {
			return false;
		}
	}
	
	public static final boolean isNumeric(String s) {
		try {
			new Integer(s);
			return true;
		} catch (Exception e) {
			return false;
		}
	}
}
